#! /usr/bin/env python
# -*- Mode: Python -*-
# -*- coding: ascii -*-

__author__     = "Steven Pettit"
__date__       = "07 Aug 2020"
__copyright__  = "Steven Pettit"
__version__    = "1.0"
__maintainer__ = "Steven Pettit"
__email__      = "kryslin@tlhea.org"
__status__     = ""
__lwver__      = "11"

import os, sys, math, traceback
import lwsdk

import lwsdk.pris
import lwsdk.pris.modeler

lwv  = lwsdk.Vector
lwp  = lwsdk.pris
lwpm = lwsdk.pris.modeler

#dot = lambda a,b : lwsdk.Vector.dot(a,b)
cross = lambda a,b : lwsdk.Vector.cross(a,b)
vmag = lambda a : lwsdk.Vector.magnitude(a)
normalize = lambda a : lwsdk.Vector.normalize(a)

class estimateFFX(lwsdk.ICommandSequence):
	def __init__(self, context):
		super(estimateFFX, self).__init__()
		#Surfaces List
		self.surf = []
		#Control List
		self.fib_dia = None
		self.fib_cnt = None
		self.fib_srf = None
		self.fib_dis = None
		self.calc = None
		self.report = None
		#Parameters
		self.dia = 0.00025
		self.cnt = 1500000
		self.srf = 1
		self.mth = 1
		self.mc = None

	def fastpolyscan(self, poly_list, poly_id):
		q = self.me.polyTag(self.me.state, poly_id, lwsdk.LWPTAG_SURF)
		if q not in poly_list: poly_list.append(q)

	def getSurfaces(self):
		rv = []
		try:
			self.me = self.mc.editBegin(0,0, lwsdk.OPSEL_USER)
			eor = self.me.fastPolyScan(self.me.state, self.fastpolyscan, (rv,), lwsdk.OPLYR_FG,0)
		except:
			print "We've had an error, closing down gracefully..."
			err = traceback.format_exc()
			print "Error : ", err 
		finally:
			self.me.done(self.me.state,lwsdk.EDERR_USERABORT,0)

		return rv

	def getFFXLength(self):
		#Get the base FiberFX fiber length
		m_bb = lwpm.boundingbox(None,False)
		bb = m_bb
		#returns 2 lists of 3 items
		ub = [bb[1][0] - bb[0][0],bb[1][1] - bb[0][1],bb[1][2] - bb[0][2]]
		d = math.sqrt(max(ub[0]**2 + ub[1]**2, ub[1]**2 + ub[2]**2, ub[2]**2 + ub[0**2]))/20.0
		return d

	def powers(self,q):
		return 2 ** q

	def squares(self,q):
		return ((2* q) + 1)**2

	def hexes(self,q):
		return 1 + 6 * ((q * (q + 1))/2)

	def fib(self,q):
		return round(((1.61803398875 ** (q+2))/math.sqrt(5)),0)

	def poly_area(self, qty, verts, pnormal):
		if qty < 3 : return 0

		area = -1
		accum = lwv(0,0,0)
		inorm = lwv(pnormal)
		for i in range(qty):
			j = (i + 1) % qty
			accum += cross(verts[i], verts[j])
				
		area = lwsdk.Vector.dot(accum,inorm)
		area /= 2
		area = abs(area)
		return area
	
	def domagic(self, id, data):
		hcr = -1
		tsa = 0.0
		erf = False

		#get values from controls
		fibwidth = self.fib_dia.get_float()
		fibcnt = self.fib_cnt.get_int()
		srfidx = self.fib_srf.get_int()
		method = self.fib_dis.get_int()

		fiblength = self.getFFXLength()

		try:
			lwpm.selmode(lwsdk.OPSEL_USER)
			lwpm.selpolygon('CLEAR','NVGT',0)
			lwpm.selpolygon('SET','SURFACE',self.surf[srfidx])

			lwpm.editbegin()
			lme = lwpm._local_me
			lp = lwpm.getpolygons()
			pc = len(lp)
			lwp.moninit(pc,"Analyzing Mesh")
			for p in lp:
				pp = lwpm.polypointinfo(p)
				pi = lwpm.polyinfo(p)
				pl = pi[1:]
				pa = []
				cp = lwv(0,0,0)
				for v in pl:
					t = lwpm.pointinfo(v)
					cp += (t / pp)
					pa.append(t)
				
				lpa = len(pa)
				for i in range(lpa): 
					hcr = max(hcr, lwsdk.Vector.magnitude(pa[i] - cp))
		
				tpa = self.poly_area( lpa, pa, lwpm.polynormal(p))
				if tpa < 0:
					print 'Error in Polygon Area!'
					erf = True
					break
				tsa += self.poly_area( lpa, pa, lwpm.polynormal(p))
				lwp.monstep(1)
			lwp.monend()
			
		except:
			print "We've had an error, closing down gracefully..."
			err = traceback.format_exc()
			print "Error : ", err 
		finally:
			lwpm.editend(lwsdk.EDERR_USERABORT)
			lme = None
		
		if erf:
			self.report.set_str('Internal Error!')
			self.report.render()
			lwpm.selpolygon('CLEAR','NVGT',0)
			return

		rcr = round(hcr * 707.10678,2)
		numfibers = min(fibcnt, math.ceil((tsa / ((.25 * math.pi) * fibwidth ** 2))/100) * 100)
		recd = 0
		recc = 0
		for i in range(11):
			if method == 0: te = self.powers(i)
			if method == 1: te = self.squares(i)
			if method == 2: te = self.hexes(i)
			if method == 3: te = self.fib(i)
			if method < 0 or method >3: te = powers(i)
			tc = te * 100000
			td = numfibers / tc
						
			if td <= 1:
				recd = math.ceil(100 * td)
				recc = te
				break

		ta = (tsa ** 0.5) * 1000
		tb = numfibers ** 0.5
		rcr = (rcr + (ta / tb)) / 2
		
		r1 = "Fiber Length   = %s mm.\r\n" % (fiblength * 1000)
		r2 = "Density        = %s percent\r\n" % (recd)
		r3 = "Cluster Fibers = %s fibers.\r\n" % (recc)
		r4 = "Cluster Radius = %s mm.\r\n" % (rcr)
		r5 = r1 + r2 + r3 + r4
		
		lwpm.selunhide()
		lwpm.selpolygon('CLEAR','NVGT',0)
		self.report.set_str(r5)
		self.report.render()

	def myUI(self):
		ctl_list = []
		self.ui    = lwsdk.LWPanels()
		self.panel = self.ui.create("Fiber FX Settings Estimator")

		self.fib_cnt = self.panel.int_ctl("Fiber Count")
		self.fib_cnt.set_int(1500000)
		ctl_list.append(self.fib_cnt)

		self.fib_dia = self.panel.dist_ctl("Fiber Diameter")
		self.fib_dia.set_float(0.000250)
		ctl_list.append(self.fib_dia)

		self.fib_srf = self.panel.popup_ctl("Fiber FX Surface",self.surf)
		ctl_list.append(self.fib_srf)

		self.fib_dis = self.panel.popup_ctl("Fiber Distribution",['Powers of 2','Grid','Hexagonal','Fibonacci'])
		ctl_list.append(self.fib_dis)

		self.fib_calc = self.panel.button_ctl("Calculate")
		self.fib_calc.set_event(self.domagic)
		ctl_list.append(self.fib_calc)
		
		self.hrule = self.panel.border_ctl("",0,0)
		ctl_list.append(self.hrule)
		
		self.report = self.panel.multiline_ctl("Results",40,10)
		self.report.set_str("Press Calculate to get Results.")
		ctl_list.append(self.report)

		self.panel.align_controls_vertical(ctl_list)
		self.panel.size_to_layout(5,5)

		if self.panel.open(lwsdk.PANF_BLOCKING | lwsdk.PANF_CANCEL) == 0:
  			self.ui.destroy(self.panel)
  			return lwsdk.AFUNC_OK # when cancel was clicked

		self.cnt = self.fib_cnt.get_int()
		self.dia = self.fib_dia.get_float()
		self.srf = self.fib_srf.get_int()
		self.mth = self.fib_dis.get_int()
		
		self.ui.destroy(self.panel)
		
	def process(self, mod_command):
		self.mc = mod_command
		lwpm.init(mod_command)
		self.surf = self.getSurfaces()
		#Everything Happens Here
		self.myUI()
		return lwsdk.AFUNC_OK
			
ServerTagInfo = [( "estimateFFX", lwsdk.SRVTAG_USERNAME | lwsdk.LANGID_USENGLISH ),
                 ( "Estimate FFX Settings", lwsdk.SRVTAG_BUTTONNAME | lwsdk.LANGID_USENGLISH )]
ServerRecord = { lwsdk.CommandSequenceFactory("Estimate FFX Settings", estimateFFX) : ServerTagInfo }