#! /usr/bin/env python
# -*- Mode: Python -*-
# -*- coding: ascii -*-

__author__     = "Steven Pettit"
__date__       = "11 Aug 2020"
__copyright__  = "Steven Pettit"
__version__    = "1.0"
__maintainer__ = "Steven Pettit"
__email__      = "kryslin@tlhea.org"
__status__     = ""
__lwver__      = "11"

#Typical includes...
import os, sys, math, traceback
import lwsdk
import lwsdk.pris
import lwsdk.pris.modeler

#abbreviations and shortcuts start here!
lwv   = lwsdk.Vector
lwsqf = lwsdk.LWStateQueryFuncs()
lwmf  = lwsdk.LWMessageFuncs()
lwp   = lwsdk.pris
lwpm  = lwsdk.pris.modeler
lwof  = lwsdk.LWObjectFuncs()

AOK = lwsdk.AFUNC_OK

USER_SEL = lwsdk.OPSEL_USER
GLOBAL_SEL = lwsdk.OPSEL_GLOBAL
DIRECT_SEL = lwsdk.OPSEL_DIRECT

USER_ABORT = lwsdk.EDERR_USERABORT
NO_ERROR = lwsdk.EDERR_NONE

VMAP_STYLE = lwsdk.LWID_('GUDE')
VMAP_WEIGHT = lwsdk.LWVMAP_WGHT

BONE  = lwsdk.LWPOLTYPE_BONE
CURVE = lwsdk.LWPOLTYPE_CURV
FACE  = lwsdk.LWPOLTYPE_FACE
MBAL  = lwsdk.LWPOLTYPE_MBAL
PATCH = lwsdk.LWPOLTYPE_PTCH
SUBD  = lwsdk.LWPOLTYPE_SUBD

clamp = lambda a, l, h : max(min(a,h),l)
sgn = lambda a : a and (1, -1)[a < 0]
lerp = lambda a,b,t : (b - a) * clamp(t,0,1) + a
lerp2d = lambda a,t,u : lerp( lerp(a[0], a[2], u), lerp(a[1], a[3], u), t)
center = lambda a,b : ( a + b ) / 2.0 

class MultitreeNode(object):
	def __init__(self, name_list):
		self.name_list = name_list
		self.flags = 0
		self.children = []

class styleComb(lwsdk.ICommandSequence):
	def __init__(self, context):
		super(styleComb, self).__init__()
		self.mc = None
		self.me = None
		self.sm = None
		self.pts = []
		self.pol = []
		self.styles = []
		self.wmaps  = []
		self.surfs  = []
		self.parts  = []

	def errMsg(self, txt1, txt2 = ''):
		if self.me: lwpm.editend()
		lwmf.error(txt1,txt2)

	def fastscan(self, a_list, a_id):
		a_list.append(a_id)

	def surfScan(self, a_list, a_id):
		q = self.me.polyTag(self.me.state, a_id, lwsdk.LWPTAG_SURF)
		if q not in a_list: a_list.append(q)

	def partScan(self, a_list, a_id):
		q = self.me.polyTag(self.me.state, a_id, lwsdk.LWPTAG_PART)
		if q not in a_list: a_list.append(q)

	def getWeights(self):
		mylist = ['(Fixed)']
		myID = VMAP_WEIGHT
		mapObj = lwof.numVMaps(myID)
		mylist.extend([lwof.vmapName(myID,i) for i in range(mapObj)])
		return mylist

	def getStyles(self):
		mylist = ['(New)']
		myID = VMAP_STYLE
		mapObj = lwof.numVMaps(myID)
		mylist.extend([lwof.vmapName(myID,i) for i in range(mapObj)])
		return mylist

	def getVMaps(self):
		#The code to get all the VMAPS - Surfaces, Weights, Parts, and Guides - goes here
		self.me = self.mc.editBegin(0,0,USER_SEL)
		self.me.fastPolyScan(self.me.state, self.surfScan, (self.surfs,),lwsdk.OPLYR_FG, 0)
		self.me.fastPolyScan(self.me.state, self.partScan, (self.parts,),lwsdk.OPLYR_FG, 0)
		self.me.done(self.me.state, USER_ABORT,0)
		self.wmaps = self.getWeights()
		self.styles = self.getStyles()
		self.surfs.sort()
		self.parts.sort()
		self.wmaps.sort()
		self.styles.sort()
		self.me = None

	def boundingBox(self):
		q = lwsqf.bbox(lwsdk.OPLYR_FG)
		r = q[1]
		return [ lwv(r[0],r[2],r[4]), lwv(r[1],r[3],r[5])]

	def pointCount(self):
		q = lwsqf.bbox(lwsdk.OPLYR_FG)
		return q[0]

	def center(self,a,b):
		return (a + b) / 2

	def getFFXLength(self):
		bb = self.boundingBox()
		ub = bb[1] - bb[0]
		d = math.sqrt(max(ub[0] ** 2, ub[1] ** 2, ub[2] **2)) / 20.0
		return d
	
	def getPntSel(self):
		flr = False
		a = lwv(0,0,0)
		la = []
		self.me = self.mc.editBegin(0,0,USER_SEL)
		self.me.fastPointScan(self.me.state, self.fastscan, (la,),lwsdk.OPLYR_FG,1)
		if len(la):
			t = len(la)
			for v in la: a += lwv(self.me.pointPos(self.me.state,v))
			a /= t
			flr = True
		self.me.done(self.me.state, USER_ABORT,0)
		self.me = None
		return [flr,a]

	def isSymOn(self):
		return bool(lwsqf.mode(lwsdk.LWM_MODE_SYMMETRY))

	def polyscan2(self, result,pid):
		q = lwpm._local_me.polyInfo(lwpm._local_me.state, pid)
		if q.numPnts == 1 or q.numPnts == 2:
			result.append(pid)

	def degenGeoCheck(self):
		result = [] 
		tmp = lwpm.lyrfg()
		fgl = tmp[0]
		tmp = lwpm.lyrempty()
		mtl = tmp[0]
		lwpm.selmode(USER_SEL)
		lwpm.selpolygon('CLEAR','NVGT',0)
		lwpm.selpolygon('SET','NVGT',0)
		lwpm.copy()
		lwpm.lyrsetfg(mtl)
		lwpm.paste()
		lwpm.mergepoints(0.000)
		lwpm.editbegin()
		lwpm._local_me.fastPolyScan(lwpm._local_me.state, self.polyscan2, (result,),lwsdk.OPLYR_FG,0)
		lwpm.editend(USER_ABORT)
		lwpm.selpolygon('CLEAR','NVGT',0)
		lwpm.selpolygon('SET','NVGT',0)
		lwpm.delete()
		lwpm.lyrsetfg(fgl)
		return (len(result) > 0)

	#UI Callbacks and helpers begin here, there are a lot of them...!
	def makeList(self, q, r):
		#makes a list from q, removes item r from list
		rv = list(r)
		if not self.selpnt: rv.pop(q)
		return rv

	#Callbacks for the multilist boxes goes here...
	def combsurf_name(self, control, userdata, row, a):
		if row == -1: return "combsurf"
		else: return self.combsurf_listNames[row]

	def combsurf_count(self, control, userdata):
		return len(self.combsurf_listNames)

	def combsurf_multi_select(self, control, user_data, row, selecting):
		if row < 0: return  # list selections are being cleared
		if selecting:
			self.sel_surfs[row] = True
		else:
			self.sel_surfs[row] = False
			
	def combpart_name(self, control, userdata, row, a):
		if row == -1: return "combpart"
		else: return self.combpart_listNames[row]

	def combpart_count(self, control, userdata):
		return len(self.combpart_listNames)

	def combpart_multi_select(self, control, user_data, row, selecting):
		if row < 0: return  # list selections are being cleared
		if selecting:
			self.sel_parts[row] = True
		else:
			self.sel_parts[row] = False

	def multi_null(self, control, a, b):
		return

	def initParms(self):
		#Initializes the parms for the brushing algorithm
		#that the UI sets.
		rv = {
			'status': 1,
			'style name': '',
			'sleek map': '',
			'sleek amt': 95.0,
			'curl map': '',
			'curl amt': 10.0,
			'grav str': 50.0,
			'bias str': 0.0,
			'usex': 0,
			'usez': 0,
			'fib map': '',
			'fib len': self.ffxl,
			'use parts': 0,
			'away':'',
			'comb': []
		}
		return rv

	def UIEvent(self, id, data):
		#code for the UI Callbacks goes here
		if id == self.styidx: 
			v = self.styidx.get_int()
			if v == 0: 
				self.stynam.unghost()
			else:
				self.stynam.ghost()
		if id == self.slkidx:
			v = self.slkidx.get_int()
			if v == 0:
				self.slkfix.unghost()
			else:
				self.slkfix.ghost()
		if id == self.crlidx:
			v = self.crlidx.get_int()
			if v == 0:
				self.crlfix.unghost()
			else:
				self.crlfix.ghost()
		if id == self.useprt:
			v = self.useprt.get_int()
			if v:
				self.awaysurf.visible(False)
				self.combsurf.visible(False)
				self.awaypart.visible(True)
				self.combpart.visible(True)
				if self.selpnt: self.awaypart.ghost()
			else:
				self.awaysurf.visible(True)
				self.combsurf.visible(True)
				if self.selpnt: self.awaysurf.ghost()
				self.awaypart.visible(False)
				self.combpart.visible(False)
		if id == self.awaysurf:
			# Need to rebuild names list in multilistbox control, and refresh
			v = self.awaysurf.get_int()
			q = self.makeList(v, self.surfs)
			self.combsurf_listNames = q
			self.sel_surfs = [False for t in q]
			self.combsurf.redraw()
		if id == self.awaypart:
			v = self.awaypart.get_int()
			q = self.makeList(v, self.parts)	
			self.combpart_listNames = q
			self.sel_parts = [False for t in q]
			self.combpart.redraw()

	def myUI(self):
		#The Code for the UI goes here...
		#This one will be a beast...
		rv = self.initParms()
		#region
		#the guts of the ui goes here...
		self.ui    = lwsdk.LWPanels()
		self.panel = self.ui.create("Style Comb for Python, Ver. 1.0")
		self.panel.setw(518)
		self.panel.seth(540)
		#styles
		self.styidx = self.panel.popup_ctl("Style", self.styles)
		self.styidx.move(76,2)
		self.styidx.set_event(self.UIEvent)
		self.stynam = self.panel.str_ctl("New Style Name", 80)
		self.stynam.move(20,29)
		#sleekness
		self.slkidx = self.panel.popup_ctl("Sleekness Control", self.wmaps)
		self.slkidx.move(14,56)
		self.slkidx.set_event(self.UIEvent)
		self.slkfix = self.panel.percent_ctl("Sleekness")
		self.slkfix.set_float(rv['sleek amt'])
		self.slkfix.move(50,83)
		#Fiber Curling
		self.crlidx = self.panel.popup_ctl("Curling Control", self.wmaps)
		self.crlidx.move(31,110)
		self.crlidx.set_event(self.UIEvent)
		self.crlfix = self.panel.percent_ctl("Curling")
		self.crlfix.set_float(rv['curl amt'])
		self.crlfix.move(67,137)
		#Gravity Strength
		self.grvstr = self.panel.percent_ctl("Gravity Strength")
		self.grvstr.move(23,164)
		self.grvstr.set_float(rv['grav str'])
		#Axis Bias Strength
		self.axbstr = self.panel.percent_ctl("Axis Bias Strength")
		self.axbstr.move(14,191)
		self.axsbx = self.panel.bool_ctl("X Axis")
		self.axsbx.move(106,218)
		self.axsbz = self.panel.bool_ctl("Z Axis")
		self.axsbz.move(312,218)
		#Fiber Length
		self.fibidx = self.panel.popup_ctl("Fiber Length Control", self.wmaps)
		self.fibidx.move(4,245)
		self.fiblen = self.panel.dist_ctl("Fiber Length")
		self.fiblen.move(40,272)
		self.fiblen.set_float(rv['fib len'])
		#Use Parts
		self.useprt = self.panel.boolbutton_ctl("Use Parts")
		self.useprt.move(106,299)
		self.useprt.set_event(self.UIEvent)
		if len(self.parts) < 2: self.useprt.ghost()
		#Surfaces Selection
		self.awaysurf = self.panel.popup_ctl("Comb Away From", self.surfs)
		self.awaysurf.move(17,326)
		self.awaysurf.set_event(self.UIEvent)
		if self.selpnt: self.awaysurf.ghost()
		q = self.makeList(0,self.surfs)
		self.combsurf_listNames = q
		self.sel_surfs = [False for t in q]
		self.combsurf = self.panel.multilist_ctl("Comb These Surfaces", 370, 10, self.combsurf_name, self.combsurf_count, self.multi_null)
		self.combsurf.move(106,353)
		self.combsurf.set_select(self.combsurf_multi_select)
		#Parts Selection
		#15
		self.awaypart = self.panel.popup_ctl("Comb Away From", self.parts)
		self.awaypart.move(17,326)
		self.awaypart.set_event(self.UIEvent)
		if self.selpnt: self.awaypart.ghost()
		# In order to add items to the list, this is how you do it, and then refresh the control:
		# just add elements to the array: self.combpart_listNames
		#16
		q = self.makeList(0,self.parts)
		self.combpart_listNames = q
		self.sel_parts = [False for t in q]
		self.combpart = self.panel.multilist_ctl("Comb These Parts", 370, 10, self.combpart_name, self.combpart_count,self.multi_null)
		self.combpart.move(106,353)
		self.combpart.set_select(self.combpart_multi_select)
		#funky UI stuff goese here...
		if self.useprt.get_int():
			self.awaysurf.visible(False)
			self.combsurf.visible(False)
			self.awaypart.visible(True)
			self.combpart.visible(True)
			if self.pnt_selected: self.awaypart.ghost()
		else:
			self.awaysurf.visible(True)
			self.combsurf.visible(True)
			if self.selpnt: self.awaysurf.ghost()
			self.awaypart.visible(False)
			self.combpart.visible(False)
		#endregion
		#region
		if self.panel.open(lwsdk.PANF_BLOCKING | lwsdk.PANF_CANCEL) == 0:
			self.ui.destroy(self.panel)
			rv['status'] = 0
			return rv # when cancel was clicked
		#endregion
		#region
		#parameter processing goes here...
		#style name
		t = self.styidx.get_int()
		if t == 0: rv['style name'] = self.stynam.get_str()
		else: rv['style name'] = self.styles[t]
		#Sleekness
		t = self.slkidx.get_int()
		rv['sleek map'] = self.wmaps[t]
		rv['sleek amt'] = (self.slkfix.get_float() * (t == 0)) / 100
		#curling
		t = self.crlidx.get_int()
		rv['curl map'] = self.wmaps[t]
		rv['curl amt'] = self.crlfix.get_float() / 100.0
		#gravity
		rv['grav str'] = self.grvstr.get_float() / 100.0
		#axis bias strength
		rv['bias str'] =self.axbstr.get_float() / 100.0
		#X axis
		rv['usex'] = self.axsbx.get_int()
		#Z axis
		rv['usez'] = self.axsbz.get_int()
		#Fiber Length Control
		t = self.fibidx.get_int()
		rv['fib map'] = self.wmaps[t]
		rv['fib len'] = self.fiblen.get_float()
		up = self.useprt.get_int()
		rv['use parts'] = up
		#Ok, now things get funky!
		tmp = []
		if up:
			#Get the various part selections
			rv['away'] = self.parts[self.awaypart.get_int()]
			q = len(self.combpart_listNames)
			for i in range(q):
				if self.sel_parts[i]: tmp.append(self.combpart_listNames[i])
			rv['comb'] = tmp
		else:
			rv['away'] = self.surfs[self.awaysurf.get_int()]
			q = len(self.combsurf_listNames)
			for i in range(q):
				if self.sel_surfs[i]: tmp.append(self.combsurf_listNames[i])
			rv['comb'] = tmp
		if self.selpnt: rv['away'] = 'point'
		#endregion
		return rv

	def vtxPolygons(self, pnt_id):
		rv = []
		q = self.me.pointInfo(self.me.state, pnt_id)
		for i in range(q.numEdges):	
			r = self.me.edgeInfo(self.me.state, q.edges(i))
			for j in range(r.numPols):
				s = r.pols(j)
				if s not in rv: rv.append(s)
		return rv		

	def makeVtxPerp(self, pt_eval, away_v):
		zv = lwv(0,0,0)
		bp = pt = tv = vn = rv = zv
		qt = self.me.pointPos(self.me.state, pt_eval)
		pt = lwv(qt)
		bp = away_v
		tv = lwv.normalize(pt - bp)
		pids = self.vtxPolygons(pt_eval)
		pc = len(pids)
		for p in pids:
			qt = self.me.polyNormal(self.me.state, p)
			qq = lwv(qt[1])
			vn +=  qq
		vn /= pc	
		vn = lwv.normalize(vn)
		rv = lwv.normalize( tv - (vn * lwv.dot(vn,tv)))
		if (pt.y + rv.y) > (pt.y + vn.y) and vn.z > 0: rv *= lwv(1,-1,1)
		if sgn(rv.x) != sgn(pt.x): rv.x = -rv.x
		return [vn, rv]

	def makeMasterBias(self, e_pt, e_vn, b_pt, t_bb, u_x, u_z):
		rv = lwv(0,0,0)
		c = lwv(0,0,0)
		l_bb = t_bb / 2.0
		
		rv.y = clamp(lwv.dot(e_vn,lwv(0,-1,0)),0,1) ** 2
		if u_x or u_z:
			c = e_pt - b_pt
			if u_x and t_bb.x !=0: 
				rv.x = abs(c.x)/l_bb.x
			if u_z and t_bb.z !=0: 
				rv.z = abs(c.z)/l_bb.z
		
		return rv

	def makePerp(self, v1, v2):
		rv = lwv(0,0,0)
		vp = v1
		vq = v2
		rv = lwv.normalize(vp - (vq * lwv.dot(vp,vq)))
		if lwv.magnitude(rv) == 0: rv = vp
		return rv

	def makeGuide(self, pnt, vn, vb, fl, bl, cl):
		tv = [lwv(0,0,0) for i in range(3)]
		gr = [pnt, pnt + (vb * fl), pnt + (vn * fl), pnt + (vn * fl) + (vb * fl)]
		nx = ny = theta = 0.0
		lb = bl
		lf = fl
		lc = cl
		theta = math.acos(clamp(lb,.1,.9))
		lcrl = 0.174533 + clamp(lc,0,1)
		tv[0] = lerp2d(gr,0,0)
		for i in range(1,3):
			nx += 0.5 * math.cos(theta)
			ny += 0.5 * math.sin(theta)
			tv[i] = lerp2d(gr,nx,ny)
			theta -= lcrl
			if theta < 0: theta += 2 * lcrl
		
		return [ tv[0].x, tv[0].y, tv[0].z,tv[1].x, tv[1].y, tv[1].z,tv[2].x, tv[2].y, tv[2].z ]

	def getPolygons(self):
		a = []
		res = self.me.fastPolyScan(self.me.state, self.fastscan, (a,),lwsdk.OPLYR_FG, 1)
		if not a and (lwpm._local_sm == USER_SEL) or (lwpm._local_sm == GLOBAL_SEL):
			res = self.me.fastPolyScan(self.me.state, self.fastscan, (a,),lwsdk.OPLYR_FG, 0)
		return a
	
	def getPoints(self):
		a = []
		res = self.me.fastPointScan(self.me.state, self.fastscan, (a,),lwsdk.OPLYR_FG, 1)
		if not a and (lwpm._local_sm == USER_SEL) or (lwpm._local_sm == GLOBAL_SEL):
			res = self.me.fastPointScan(self.me.state, self.fastscan, (a,),lwsdk.OPLYR_FG, 0)

	def _slowscan(self, a_list, a_id):
		if a_id.flags & lwsdk.EDDF_SELECT:
			for i in range (a_id.numPnts):
				if a_id.getpoint(i) not in a_list: a_list.append(a_id.getpoint(i))

	def getPointsFromPolys(self):
		a = []
		res = self.me.polyScan(self.me.state, self._slowscan,(a,),lwsdk.OPLYR_FG)
		return a

	def bbox2(self):
		la = []
		hi = [-999999999999.99,-999999999999.99,-999999999999.99]
		lo = [999999999999.99,999999999999.99,999999999999.99]
		try:
			lme = self.mc.editBegin(0,0,USER_SEL)
			res = lme.fastPolyScan(lme.state, self.fastscan,(la,),lwsdk.OPLYR_FG,1)
			if not la:
				res = lme.fastPolyScan(lme.state, self.fastscan,(la,),lwsdk.OPLYR_FG,0)
			for p in la:
				q = lme.polyInfo(lme.state, p)
				for i in range(q.numPnts):
					r = lme.pointInfo(lme.state, q.getpoint(i))
					if r.position[0] < lo[0] : lo[0] = r.position[0]
					if r.position[0] > hi[0] : hi[0] = r.position[0]
					if r.position[1] < lo[1] : lo[1] = r.position[1]
					if r.position[1] > hi[1] : hi[1] = r.position[1]
					if r.position[2] < lo[2] : lo[2] = r.position[2]
					if r.position[2] > hi[2] : hi[2] = r.position[2]
				
		except:
			print "We've had an error, closing down gracefully..."
			err =  traceback.format_exc()
			print "Error : ", err
		finally:
			lme.done(lme.state,USER_ABORT,0)
	
		return [lwv(lo),lwv(hi)]

	def process(self, mod_command):
		#Store the mod_command and initialize PRIS.
		#Why use PRIS?  Simpler Syntax to get things done, mainly.
		#It does, however, have issues.  getpoints/polygons doesn't respect selmode, for instance
		self.mc = mod_command
		if not lwpm.init(self.mc):
			lwmf.error('Failed to initialize PRIS Module.','')
			return AOK
		#We need to exit if there's no geometry...
		ge = self.pointCount()
		if not ge:
			lwmf.error('Need some geometry to work with.','Load something and start again.')
			return AOK
		#Let's grab our vertex maps
		self.getVMaps()
		#Set the default fiber length
		self.ffxl = self.getFFXLength()
		#Get any point selections made
		tmp = self.getPntSel()
		self.selpnt = tmp[0]
		self.brushp = tmp[1]
		#Check for symmetry, exit if on.
		if self.isSymOn():
			lwmf.error('Symmetry is on.','Please disable it and re-run plugin.')
			return AOK
		#Check for 1-2pt polygons, exit if they exist.
		if self.degenGeoCheck():
			lwmf.error('Degenerate geometry (1 & 2 pt polygons) are present.','Please delete them and re-run plugin.')
			return AOK
		#All checks required before starting are complete!
		
		try:
			#Generate our UI and get our parameters!
			parms = self.myUI()
			if parms['status'] == 1:
			#We pressed OK, let's do it, and hopefully not crash lightwave to the desktop!
			#Let's get our brushing point.  All combing is away from here.
			# A slightly cleaner solution than if...else blocks. parms['use parts'] is either 0 or 1
				tcmd = ['SURFACE','PART']
				if not self.selpnt:
					lwpm.selpolygon('CLEAR','NVGT',0)
					lwpm.selpolygon('SET', tcmd[parms['use parts']], parms['away'])
					qt = self.bbox2()
					if not qt:
						lwmf.error('First Bounding Box Call Failed.','')
						return AOK
					self.brushp = center(qt[0],qt[1])
				
				#Now, let's get our selection to be combed...
				lwpm.selpolygon('CLEAR','NVGT',0)
				for s in parms['comb']: lwpm.selpolygon('SET', tcmd[parms['use parts']],s)
				rt = self.bbox2()
				if not rt:
					lwmf.error('Second Bounding Box Call Failed.','')
					return AOK
				bbs = rt[1] - rt[0]
				cps = center(rt[0],rt[1])
				#Ok, let's move the brushp based on distance from center of the last box
				if not self.selpnt:
					tp1 = lwv(self.brushp.x,self.brushp.y,qt[0].z)
					tp2 = lwv(self.brushp.x,self.brushp.y,qt[1].z)
					if lwv.magnitude(cps - tp1) > lwv.magnitude(cps - tp2):
						self.brushp = tp1
					else:
						self.brushp = tp2
				#Debugging starts here...
				lwpm.undogroupbegin()
				lwpm.selmode(USER_SEL)
				lwpm.editbegin()
				self.me = lwpm._local_me
				fl_fix =  True
				sl_fix =  True
				cl_fix =  True
				sm_id = fl_id = sl_id = cl_id = None
				sm_id = self.me.vMapSelect(self.me.state, parms['style name'], VMAP_STYLE, 9)
				if not sm_id:
					self.errMsg('Unable to instance style map.')
					return AOK
				if parms['fib map'] != '(Fixed)':
					fl_id = self.me.vMapSelect(self.me.state, parms['fib map'], VMAP_WEIGHT,1)
					if not fl_id:
						self.errMsg('Unable to instance fiber length map.')
						return AOK
					else:
						fl_fix = False
				if parms['sleek map'] != '(Fixed)':
					sl_id = self.me.vMapSelect(self.me.state, parms['sleek map'], VMAP_WEIGHT,1)
					if not fl_id:
						self.errMsg('Unable to instance sleekness control map.')
						return AOK
					else:
						sl_fix = False
				if parms['curl map'] != '(Fixed)':
					cl_id = self.me.vMapSelect(self.me.state, parms['curl map'], VMAP_WEIGHT,1)
					if not cl_id:
						self.errMsg('Unable to instance curling control map.')
						return AOK
					else:
						cl_fix = False
				points = self.getPointsFromPolys()
				tv = lwv(0,0,0)
				if not points:
					self.errMsg('We\'ve had a slight problem.','Please stand by...')
					return AOK
				
				lwp.moninit(len(points), 'Working','Working...')
				for vtx in points:
					p = self.vtxPolygons(vtx)
					if len(p) > 0:
						nt = self.me.pointPos(self.me.state, vtx)
						pt = lwv(nt)
						tm = self.makeVtxPerp(vtx, self.brushp)
						vn = tm[0]
						vb = tm[1]
						tb = lwv.normalize(pt - self.brushp)
						mbias = self.makeMasterBias( pt, vn, self.brushp, bbs, parms['usex'], parms['usez'])	
						tv.x = sgn(pt.x - self.brushp.x) * mbias.x
						tv.z = sgn(pt.z - self.brushp.z) * mbias.z
						vb += (tv * parms['bias str'])
						vn = lwv.normalize(vn)
						vb = lwv.normalize(vb)
						vb = self.makePerp(vb,vn)
						#Fur Length
						fib_len = parms['fib len']
						blendst = parms['sleek amt']
						crl_amt = parms['curl amt']
						if not fl_fix:
							self.me.pointVset(self.me.state, fl_id, VMAP_WEIGHT, parms['fib map'])
							res = self.me.pointVGet(self.me.state, vtx)
							if res[0]: fib_len = parms['fib len'] * res[1]
							else: fib_len = 0
						if not sl_fix:
							self.me.pointVset(self.me.state, sl_id, VMAP_WEIGHT, parms['sleek map'])
							res = self.me.pointVGet(self.me.state, vtx)
							if res[0]: blendst = res[1]
							else: blendst = 0
						if not cl_fix:
							self.me.pointVSet(self.me.state, cl_id, VMAP_WEIGHT, parms['curl map'])
							res = self.me.pointVGet(self.me.state, vtx)
							if res[0]: crl_amt = res[1]
							else: crl_amt = parms['curl amt']
						blendst *= (1 - mbias.y) 
						crl_amt *= (1 - mbias.y)
						#make our guide
						#this will return a list of nine floats
						guide = self.makeGuide( pt, vn, vb, fib_len, blendst, crl_amt)
						#Ok, now to write the guide to the vertex map...
						self.me.pointVSet(self.me.state, sm_id, VMAP_STYLE, parms['style name'])
						self.me.vMapSet(self.me.state, vtx, None, guide)
						guide = []
						lwp.monstep()

				lwp.monend()
				lwpm.editend()
				self.me = None
				lwpm.undogroupend()
		except:
			print "We've had an error, closing down gracefully..."
			err =  traceback.format_exc()
			print "Error : ", err
		finally:
			if self.me :
				lwpm.editend()
				lwpm.undogroupend()
			lwpm.selpolygon('CLEAR','NVGT',0)

		return AOK
			
ServerTagInfo = [( "styleComb", lwsdk.SRVTAG_USERNAME | lwsdk.LANGID_USENGLISH ),
                 ( "Style Comb", lwsdk.SRVTAG_BUTTONNAME | lwsdk.LANGID_USENGLISH )]
ServerRecord = { lwsdk.CommandSequenceFactory("Style Comb", styleComb) : ServerTagInfo }